/* Open SoundControl kit in C++                                              */
/* Copyright (C) 2002 Daniel Holth, Clinton McChesney                        */
/*                                                                           */
/* This library is free software; you can redistribute it and/or             */
/* modify it under the terms of the GNU Lesser General Public                */
/* License as published by the Free Software Foundation; either              */
/* version 2.1 of the License, or (at your option) any later version.        */
/*                                                                           */
/* This library is distributed in the hope that it will be useful,           */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of            */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU         */
/* Lesser General Public License for more details.                           */
/*                                                                           */
/* You should have received a copy of the GNU Lesser General Public          */
/* License along with this library; if not, write to the Free Software       */
/* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */
/*                                                                           */
/* For questions regarding this program contact                              */
/* Daniel Holth <dholth@stetson.edu> or visit                                */
/* http://www.stetson.edu/~ProctoLogic/                                      */

/**
 * @class QueuedBroadcaster
 *
 * Holds messages in a queue, awaiting transmission by a helper class.
 *
 * @author Daniel Holth
 * @author Clinton McChesney
 */

#ifndef QUEUED_BROADCASTER_H
#define QUEUED_BROADCASTER_H

#include "Broadcaster.h"
#include <semaphore.h>
#include <queue>
#include <utility>

class QueuedBroadcaster : public Broadcaster
{
 protected:
  typedef std::queue< std::pair<Transmit*, std::string> > MessageQueue;

 private:
  MessageQueue     myMessageQueue;
  sem_t*           mySemaphore;     ///< Global count of queued messages
  pthread_mutex_t  myMutex;
    
 protected:
  MessageQueue*    getMessageQueue() { return &this->myMessageQueue; }
  sem_t*           getSemaphore()    { return this->mySemaphore;     }
  pthread_mutex_t* getMutex()        { return &this->myMutex;        }
    
  /// Enqueue @a data for later transmission.
  /** If there are subscribers, we enqueue @a data,
   * if there are not, we discard it.
   */
  virtual bool transmitInternal(Transmit* destination, std::string *data);
    
 public:
  QueuedBroadcaster();
   
  /// Send the next outgoing message.
  /** @return true if a message was sent, false if queue is empty.
   */
  bool sendOne();

  /// Returns true if the message queue is empty.
  bool empty();

  void setSemaphore(sem_t* sem) { this->mySemaphore = sem; }

  /** get name (specializes getName in Transmit)
   *
   * \return a string which identifies this object (incl. type).
   */
  virtual std::string getName() const = 0;

};

#endif
